local lg = love.graphics
local lw = love.window
local old_stencil = lg.stencil


local white = function() lg.setColor(255,255,255) end
local main  = {075,075,105}
local black = function() lg.setColor(000,000,000) end
local clred = {255 ,50,0,255}
local degrees = 0
local rotated = function() lg.setColor(rotate_hue(clred, degrees, -degrees)) end


---------- 0.10 to 0.11 -------------
local _major, _minor, _revision, _codename = love.getVersion()

if _major>=11 then
   local old = love.graphics.setColor
   function lg.setColor(arg,two,three,four)
      if two then -- there's 3 or 4 vars
         old(arg/255,two/255,three/255, (four or 255)/255)
      else -- arg is a table
         if arg then
            old(arg[1]/255,arg[2]/255,arg[3]/255, (arg[4] or 255)/255)
         else
            local arg = main
            old(arg[1]/255,arg[2]/255,arg[3]/255, (arg[4] or 255)/255)
         end
      end
   end

   local _old = lg.getColor
   function lg.getColor()
      local ret = {_old()}
      return ret[1]*255,ret[2]*255,ret[3]*255,ret[4]*255
   end

   lw.getPixelScale = lw.getDPIScale
end

_major, _minor, _revision, _codename = nil, nil, nil, nil
-------------------------------------

io.stdout:setvbuf("no")

local utf8 = require "utf8"

local function utf8sub(str, startChar, numChars)
  local startIndex = 1
  while startChar > 1 do
      local char = string.byte(str, startIndex)
      startIndex = startIndex + chsize(char)
      startChar = startChar - 1
  end

  local currentIndex = startIndex

  while numChars > 0 and currentIndex <= #str do
    local char = string.byte(str, currentIndex)
    currentIndex = currentIndex + chsize(char)
    numChars = numChars -1
  end
  return str:sub(startIndex, currentIndex - 1)
end

local sign = function(num)
   local res = 0
   if num>=0 then res = 1 else res = -1 end
   return res
end


function acronum(num)
   local nums = math.modf(math.log10(num))
   local affix = ""
   local div = 1
   if nums >2  then affix = "K" div = 3  end -- kilo
   if nums >5  then affix = "M" div = 6  end -- mega
   if nums >8  then affix = "G" div = 9  end -- giga
   if nums >11 then affix = "T" div = 12 end -- tera
   if nums >14 then affix = "P" div = 15 end -- peta
   if nums >17 then affix = "E" div = 18 end -- exo

   if nums>2 then
      num = num/10^(div)
   end

   return string.sub(num,1,5)..affix
end


lg.last_stencil = nil
function lg.stencil(...)
   lg.last_stencil = {...}
   old_stencil(...)
end


local GUI = {}
GUI.controls = {}
GUI.wheelmoved = {0, 0}
GUI.draw_level = 0
GUI.scale = 1
GUI.focused_input = nil
GUI.sin_timer = 0
GUI.width, GUI.height = lg.getDimensions( )
if love.system.getOS()=="Android" then 
   if GUI.width< GUI.height then GUI.width, GUI.height = GUI.height, GUI.width end
   SWAPPED = 'SWAPPED: true'
end
GUI.bgColor = {0.3,0.3,0.3}
GUI.rounded_corners = 0

function GUI:set_rounded_corners(num)
   GUI.rounded_corners = num
end


function GUI:set_main(f)
   main = f
end

function GUI:get_main()
   return main
end


GUI.text_buff = ""

love.keyboard.setKeyRepeat(true)

function love.textinput(t)
    if GUI.focused_input then
       if utf8.len(GUI.text_buff)< (GUI.focused_input.max or 10000) then
          GUI.text_buff = GUI.text_buff .. t
          GUI.focused_input:setText(GUI.text_buff)
       end
    end
end

function GUI:resetTextBuffer()
   GUI.text_buff = ''
end

function love.keypressed(key,unicode, repeats)
    if GUI.focused_input then
       local caretposx = GUI.focused_input.caretx
       if key == "backspace" then
          local byteoffset = utf8.offset(GUI.text_buff, -1)
          if byteoffset then
             GUI.text_buff = string.sub(GUI.text_buff, 1, byteoffset - 1)
             GUI.focused_input:setText(GUI.text_buff)
          end
       elseif key=="space" then
          key = " "
       elseif key=="tab" then
          if GUI.focused_input then
             GUI.focused_input = GUI.focused_input.next
             GUI.text_buff = (GUI.focused_input.text or "")
          end
          --GUI.focused_input_caret  = utf8.len(GUI.focused_input.text or "")
       elseif key == "left" then
            --  if GUI.focused_input_caret > 1 then
             --    GUI.focused_input_caret =  GUI.focused_input_caret - 1
             -- end
       elseif key == "right" then
              --if GUI.focused_input_caret < utf8.len(GUI.focused_input.text or "") then
              --   GUI.focused_input_caret =  GUI.focused_input_caret + 1
              --end
       end
    end
end
local str = {}


function GUI:remove_control(control)
   self:_remove_control(self.root,control)
end

function GUI:_remove_control(rt,control)
   for i, child in pairs(rt.children or {}) do -- iterate children only
       if child and control and child==control then
          --print("found as child of "..tostring(rt).." ("..rt.type..")",control)
          rt.children[i] = nil
          --print("there are {"..tostring(#rt.children).."} children left")
          for i=1, #rt.children do
              --print(rt.children[i])
          end
          --break
       else
          self:_remove_control(child, control)
       end
   end
end

--[[
function GUI:remove_control(control)
   if not control then return end
   local add =" (parentless)"
   if control.parent then
      add = " on "..control.parent.type.."[".. control.parent.id.."]"
   end
   print("Checking for "..control.type.."[".. control.id.."]"..add)
   self:_remove_control(self.root,control,0)
end

function GUI:_remove_control(rt,control,depth)
   depth = depth + 1
   for i, child in ipairs(rt.children or {}) do -- iterate children only
          --print("--- checking ".. child.type.."[".. child.id.."]'s children" )
       print(string.rep("\t",depth).."L_ "..child.type.."[".. child.id.."] (as a child of the ".. rt.id .."), serching "..control.id)

       if child == control then
          if child.children and #child.children>0 then
             for j=1, #child.children do
                 self:_remove_control(child, child.children[j],depth)
             end
          end
          print(string.rep("\t",depth).."rm ".. control.type.."[".. control.id.."]" )
          control:setParent()
          table.remove(self.controls,control.id)
          for idx = 1, #self.controls do
              if self.controls[idx] == control then
                 print(control.id .. " still exists")
                 table.remove(self.controls,idx)
              end
          end
          print(string.rep("\t",depth).."rm ".. control.type.."[".. control.id.."]" )

          --self:_remove_control(control.parent,control,depth)
          control = nil
          child = nil
          break
       else
          if child.children and #child.children>0 then
             for j=1, #child.children do
                 self:_remove_control(child, control,depth)
             end
          end
       end
   end
   depth = depth - 1
end
]]

function GUI:_draw(control)
  -- get self.controls and tvarse that instead
  -- also check if a control was drawn already via some flag
   if not (control.visible and (control.parent or control==self.root)) then return end
   local comparemode, comparevalue = lg.getStencilTest()
   str[#str+1] = comparevalue
   lg.setStencilTest("gequal", self.draw_level)
   
   local font = lg.getFont()
   lg.setFont(control.font or GUI.font)   
   control:onDraw()
   lg.setFont(font)
   --print(control.type,control.id)

   local stencil = function() lg.rectangle("fill", control.x, control.y, control.w, control.h, control.rounded_corners or GUI.round_corners) end
   if control.debug_bounds then
      --print(control.type, control.x, control.y, control.w, control.h)
      lg.setColor(55,255,math.random(255))
      lg.rectangle("line", control.x, control.y, control.w, control.h, control.rounded_corners or GUI.round_corners)
   end

   local _____last_stencil
   if control.topmost then
      _____last_stencil = self.draw_level
      self.draw_level = 999999
      lg.stencil(stencil, "replace", 999999, true)
   else
      lg.stencil(stencil, "increment", 1, true)
      lg.setStencilTest("gequal", self.draw_level)
      self.draw_level = self.draw_level + 1
   end

   table.sort(control.children, function( a,b ) return a.id <= b.id end)
   for i, child in ipairs(control.children) do -- iterate children only
       self:_draw(child)
   end

   if control.topmost then
      self.draw_level = _____last_stencil
      lg.stencil(stencil, "replace", _____last_stencil, true)
   else
      self.draw_level = self.draw_level - 1
      lg.stencil(stencil, "decrement", 1, true)
   end

end

function GUI:draw(x,y)
   if self.bgColor then
      lg.clear(self.bgColor)
   end

   lg.setStencilTest()
   x = x or 0
   y = y or 0
   lg.translate(x,y)
   str = {}
   lg.scale(self.scalex or self.scale, self.scaley or self.scale)
   table.sort(self.root.children, function( a,b ) return a.id < b.id end)
   self:_draw(self.root, str)
   --print(table.concat(str))
   lg.setStencilTest()
end

GUI.hitTest = function (xx,yy, x,y,w,h)
   return (xx>=x and xx<=x+w and yy>=y and yy<=y+h) or false
end

function GUI:setScale(scale)
   self.scale = scale
end


function GUI:update(dt)
   self.sin_timer = GUI.sin_timer + dt * 10

   self.root.w = self.width
   self.root.h = self.height
   if love.system.getOS()=="Android" then 
      if GUI.width< GUI.height then 
         self.root.w = self.height
         self.root.h = self.width
      end
      SWAPPED = 'SWAPPED: true x2'
   end

   local touches = {}


   if love.mouse then
      touches[1] = {x=love.mouse.getX()/self.scale, y=love.mouse.getY()/self.scale, buttons = {love.mouse.isDown(1),love.mouse.isDown(2),love.mouse.isDown(3)}}
   end
   --[[
   if love.touch then
      local touches = love.touch.getTouches()
      for i, touch in ipairs(touches) do
          local x, y = love.touch.getPosition(touch)
          touches[#touches+1] = {x=x/GUI.scale, y=y/GUI.scale, buttons={true}}
      end
   end]]

   self.touches = touches

   local visible_controls = {}
   local toremove = {}

   local noob = #self.controls
   local upd = false
   for i=1, noob do
       if self.controls[i].destroy then
          self.controls[i] = nil
          upd = true
       end
   end
   --print('upd', upd)

   if upd then
      local justice = 0
      for i=1, noob do
          if self.controls[i]~=nil then
             justice = justice + 1
             self.controls[justice]=self.controls[i]
          end
      end

      for i = justice+1, noob do
          self.controls[i] = nil
      end

      if noob ~= #self.controls then
         --os.exit()
      end
   end

   for i,control in ipairs(self.controls) do
       if control.id~=i then
          control.id= i
       end
   end
   


   local touch = touches[1]

   local topmostcontrol = self.root
   for i,control in ipairs(self.controls) do
       if control.parent and control.visible then
          if control.update then
             control:update(dt)
          end
          visible_controls [#visible_controls+1] = control
       end

       --control.hovered_pos = {touch.x, touch.y}
       --control.hovered = true
   end
   
   --print('visible', #visible_controls)

   -- now we have visible controls
   -- we need select the one that's on the top
   -- for that we iterate all visible and hittest them
   -- the ones passed the test are the ones under touch (touches[1] only)
   -- those passed form yet another list with only touched controls
   local hitcontrols = {}
   --local ss = ""
   local scrolls = {}
   for i, control in ipairs(visible_controls) do
       -- костыль
       local hit = nil
       local cont = control.parent
       while cont do
           if self.hitTest(touch.x, touch.y, cont.x, cont.y, cont.w, cont.h) then
              hit = true
           else
              hit = false
              break
           end
           cont = cont.parent
       end


       if self.hitTest(touch.x, touch.y, control.x, control.y, control.w, control.h) then
          if hit or control.topmost then
             hitcontrols[#hitcontrols+1] = control
             control.hovered_pos = {touch.x, touch.y}
             control.hovered = true
             --print(control.name, control.type)
             if control.type == "TScrollBox" then
                scrolls[#scrolls+1] = control
             end
          end
          --ss = ss ..  (control.type or "nil") .. ", "
       end
   end


   --print(ss)

   -- inside there we should check for the zindex
   -- the higher - the better
   local topmostcontrol = hitcontrols[1] or self.root
   for i, control in ipairs(hitcontrols) do
       if topmostcontrol.id<control.id then
          topmostcontrol =  control
       else
          control.hovered_pos = nil
       end
   end

   control = topmostcontrol
   --print("TopMost",control.type)

   if control.onHover then
      control:onHover(touch.x, touch.y)
      if control.parent and control.parent.type=="TScrollBox" then
         control.parent:onHover(touch.x, touch.y)
      end
   end

   for i, box in ipairs(scrolls) do
       if box == topmostcontrol.parent then
          if self.wheelmoved[1] ~= 0 or self.wheelmoved[2]~=0 then
             if box.onMouseWheel then
                box:onMouseWheel(self.wheelmoved[1], self.wheelmoved[2])
             end
          end
       end
   end

   --if control.type~="TScrollBox" then
          if self.wheelmoved[1] ~= 0 or self.wheelmoved[2]~=0 then
             if control.onMouseWheel then
                control:onMouseWheel(self.wheelmoved[1], self.wheelmoved[2])
             end
          end
   --end


   for i,cont in ipairs(self.controls) do
       if not (cont==control) then
          -- print("leave")
          if cont.onMouseLeave then
             cont:onMouseLeave(touch.x, touch.y)
          end
       end
   end
      --print(control.type, control.pressed, control.hovered, control.pressed_last)
      if control.onClick then
                --  NIL                                   NIL
         if not control.pressed and control.hovered and control.pressed_last then
            if ( self.strict_clicks
            and touch.x == control.pressed_last[1]
            and touch.y == control.pressed_last[2])

            or not self.strict_clicks then
                control:onClick(control.pressed_last[1], control.pressed_last[2], control.pressed_last[3])

                if control.parent and control.parent.type=="TScrollBox" then
                 control.parent:onClick(control.pressed_last[1], control.pressed_last[2], control.pressed_last[3])
                end
            end
            control.pressed_last = nil
           -- print("plat",control.pressed_timer)
         end
      end



      if control.onMouseDown then
          for i,touch in ipairs(touches) do
              if control.hovered then
                 local pressed = false
                 if touch.buttons[1] or touch.buttons[2] or touch.buttons[3] then
                    if self.hitTest(touch.x, touch.y, control.x, control.y, control.w, control.h) then
                       control:onMouseDown(touch.x, touch.y, touch.buttons)
                       if control.parent and control.parent.type=="TScrollBox" then
                          control.parent:onMouseDown(touch.x, touch.y, touch.buttons)
                       end
                       --print("test",control.pressed_timer)
                       if (control.pressed_timer or 0)>0.5 then
                          if control.onLongTouch then
                             control:onLongTouch(touch.x, touch.y, touch.buttons)
                             --control.pressed_timer = nil
                             --pressed = false
                             --control.pressed_last = nil
                             --control.pressed_last2 = nil
                       -- HERE IS THE LIMITER!!!
                             --break
                          end
                       end
                       control.pressed_last = {touch.x, touch.y, touch.buttons}
                       control.pressed_last2 = {touch.x, touch.y, touch.buttons}
                       pressed = true
                    end
                 end
                 if pressed then
                -- print("pres",control.pressed_timer)
                    control.pressed_timer = (control.pressed_timer  or 0 ) + dt
                 else
                    control.pressed_timer = 0
                    --print("rese",control.pressed_timer)
                 end
                 if control.pressed and control.hovered and not pressed then
                    control:onMouseUp(0,0,{false,false,false})
                    control.pressed_last2 = nil
                   -- print("pla2",control.pressed_timer)
                 end
                 control.pressed = pressed
              end
          end
       end

   self.wheelmoved[1], self.wheelmoved[2] = 0, 0
end

function love.wheelmoved(x,y)
   GUI.wheelmoved[1], GUI.wheelmoved[2] = x, y
end


----------------------------------------------------------------------
function inherited(control, method, ...)
   if control.ancestor and self.available_controls[control.ancestor] then

      local ancestor = self.available_controls[control.ancestor]
      if ancestor[method] then
         ancestor[method](control,...)
      end
   end
end

-- TControl ----------------------------------------------------------
--local TControl = {bgColor = {075,075,105, 180}, fontColor={255,255,255}, type = "TControl"}
local TControl = {bgColor = {220,220,220, 255}, fontColor={35,35,35,255}, type = "TControl"}

TControl.__index = TControl

setmetatable(TControl, {
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})

function TControl:free() self.destroy = true end

function TControl:_init(x,y,w,h, init_data)
  self.x  = x
  self.y  = y
  self.w  = w
  self.h  = h
  self.text_align = "left"
  self.type = "TControl"
  self.children = {}
  if init_data then
     for k,v in pairs(init_data) do
        self[k] = v
     end
  end
  if self.parent then
     self:setParent(self.parent)
  end
  if self.root then
     self:setParent(GUI.root)
  end
  self.visible=false -- why&
  table.insert(GUI.controls,self)
  self.id = #GUI.controls
  return self
end

function TControl:setFontColor(col)
   self.fontColor = col
   return self
end


function TControl:setParent(control) -- control is a passed parent-to-be control. If none supplied, self.parent is checked.
   if not control then
      if self.parent then
         -- manage old parent
         table.remove(self.parent.children, self.id)
         --print("deleted parent")
         return
      end
   end
   -- set new parent
   local control = control or self.parent
   self.parent = control
   self.basex = self.x + self.parent.x
   self.basey = self.y + self.parent.y
   self.x = self.x + self.parent.x
   self.y = self.y + self.parent.y
   self.zindex = (self.parent.zindex or 0)+1

   table.insert(self.parent.children, self)
   return self
end

function TControl:hide()
   self.f = function(control) control.visible = false for k,v in ipairs(control.children) do self.f(v) end end
   self.f(self)
   return self
end

function TControl:show(test)
   self.f = function(control) control.visible = true for k,v in ipairs(control.children) do self.f(v) end end
   self.f(self)
   return self
end

function TControl:setFocus()
   GUI.focused_input = self
end

local warning = [[ I don't wanna fix this. Just let these things be. ]]

function TControl:setTransparency(value) self.transparency=value return self end
function TControl:onMouseDown() return self end
function TControl:onMouseUp() return self end
function TControl:onMouseLeave() return self end
function TControl:onClick() return self end
function TControl:onLongTouch() return self end
function TControl:onHover() return self end
function TControl:onDraw() if not self.visible then return end return self end
function TControl:setPosition(x,y) self.x = x or self.x self.y = y or self.y return self end
function TControl:setSize(w,h) self.w = w or self.w self.h = h or self.h return self end

----------------------------------------------------------


------ TScrollBox ----------------------------------------
local TScrollBox = {ancestor = "TControl", type = "TScrollBox"}
TScrollBox.__index = TScrollBox

setmetatable(TScrollBox, {
  __index = TControl, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})

function TScrollBox:_init(...)
  TControl._init(self, ...) -- call the base class constructor
  self.type = "TScrollBox"
  self.scrollpos = {0, 0}
  self.movingx = self.scrollpos[1]
  self.movingy = self.scrollpos[2]
end


function TScrollBox:update(dt,...)
   local test_bottom = 0
   local test_right  = 0
   local trueheight = 0
   for _, control in ipairs(self.children) do
       test_bottom = control.basey + control.h + 0--4

       --print("conytols",_,control.basey,self.basey,math.max(self.basey,control.basey))
       if test_bottom > trueheight then
          trueheight = test_bottom
       end
       test_right =  control.x + control.w + 0--4
       if test_right > self.w then
          --self.w = test_right
       end
   end
   self.trueheight = trueheight
   --print(trueheight)
   if self.movingy then
      self.scrollpos[2] = self.scrollpos[2] + self.movingy * dt*70
      if self.scrollpos[2] > 0 then self.scrollpos[2] = 0 end
      if self.scrollpos[2] < -self.h then end
      self.movingy = self.movingy - self.movingy*dt*10
   end


   if self.movingx then
      self.scrollpos[1] = self.scrollpos[1] + self.movingx * dt * 70
      self.movingx = self.movingx - self.movingx*dt *10
   end

   if math.abs(self.movingy)<1 then self.movingy = 0 end
   if math.abs(self.movingx)<1 then self.movingx = 0 end

   local percent = -self.scrollpos[2]/(trueheight-self.h-self.basey+0)
   if percent< 0 then self.scrollpos[2] = 0 end
   if percent> 1 then self.scrollpos[2] = -trueheight+self.h+self.basey-0 end

   self.knobposy = (self.h-40)*percent

   if percent>1 then
      self.knobposy = (self.h-40)
   end
   if percent<0 then
      self.knobposy = 0
   end


   local delta = self.y - self.parent.y - self.scrollpos[2]
      --self.y = self.parent.y + self.scrollpos[2]
   --if delta<self.basey then delta = self.basey end
   --print(self.scrollpos[2])
   self:update_children_pos()

   return self
end

function TScrollBox.update_children_pos(control)
   for i, child in ipairs(control.children) do
       child.y = child.basey+control.scrollpos[2]
       control.update_children_pos(child)
       --child:setText(child.x..":"..child.y)
   end
   return self
end

function TScrollBox:onMouseDown(x,y, buttons)
   if (x>self.w+self.x-30)  and ((self.trueheight or 0)> self.h) then
      local percent = (y-self.y)/(self.h)
      self.scrollpos[2] = -percent*(self.trueheight-self.h-self.basey+0)
      self.knobposy = (self.h)*percent
   else
      if not self.finger_down then
         self.finger_down = y
      end
   end
   return self
end

function TScrollBox:onHover()
   if self.finger_down  and ((self.trueheight or 0)> self.h) then
      local y= math.max(self.y,love.mouse.getY()/GUI.scale)
      local swipe_length = self.finger_down - y
      self.movingy = self.movingy - swipe_length/7
      self.finger_down = nil
   end
   return self
end

function TScrollBox:onMouseWheel(x,y)
   --print(self.name)
   if self.hovered and ((self.trueheight or 0)> self.h) then
      self.movingy = y*20
      self.movingx = x*20
   end
   return self
end

function TScrollBox:onDraw()
   if not TControl.onDraw(self) then return end

   if self.fill then
      lg.setColor(self.bgcol or self.bgColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("fill", self.x, self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
   end

   if self.outlineColor then
      lg.setColor(self.outlineColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("line", self.x, self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
   end

   lg.setColor(self.fontColor)
   if ((self.trueheight or 0)> self.h) then
      lg.rectangle("line", self.x + self.w-20, self.y+20, 1, self.h-40, self.rounded_corners or GUI.rounded_corners)
      lg.setColor(self.fontColor)
      lg.rectangle("fill", self.x + self.w-30, self.y+(self.knobposy or 0)+10, 20, 20, self.rounded_corners or GUI.rounded_corners)
      lg.setColor(self.bgColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("line", self.x + self.w-30, self.y+(self.knobposy or 0)+10, 20, 20, 8)
      lg.rectangle("line", self.x + self.w-30+1, self.y+(self.knobposy or 0)+10+1, 20-2, 20-2, 8)
   end
end




----- TScrollbar ---------------------------------------------

local TScrollBar = {ancestor = "TControl", type = "TScrollBox"}
TScrollBar.__index = TScrollBar

setmetatable(TScrollBar, {
  __index = TControl, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})

function TScrollBar:_init(...)
  TControl._init(self, ...) -- call the base class constructor
  self.type = "TScrollBar"
  self.scrollpos = {0, 0}
  self.movingx = self.scrollpos[1]
  self.movingy = self.scrollpos[2]
  self.position = 1
end

function TScrollBar:setOrientation(orientation)
   self.orientation = orientation
   return self
end

function TScrollBar:onMouseDown(x,y, buttons)
   local percent
   if self.orientation == "vertical" then
      percent = (y-self.y-self.w/2)/(self.h-self.w)
      if percent<0 then percent = 0 end
      if percent>1 then percent = 1 end
      self.scrollpos[2] = -percent*(self.h-self.basey+5)
      self.knobpos = (self.h-self.w)*percent
   else
      percent = (x-self.x-self.h/2)/(self.w-self.h)
      if percent<0 then percent = 0 end
      if percent>1 then percent = 1 end
      self.scrollpos[2] = -percent*(self.w)
      self.knobpos = (self.w-self.h)*percent
   end
   self.position = percent
   return self
end

function TScrollBar:update()

end

function TScrollBar:setPosition(value)
   self.position = value
   if self.orientation == "vertical" then
      self.knobpos = (self.h-self.w)*value
   else
      self.knobpos = (self.w-self.h)*value
   end
   return self
end

function TScrollBar:getPosition()
   return self.position
end


function TScrollBar:onHover()
   if self.finger_down then
      if self.orientation == "vertical" then
         local y= math.max(self.y,love.mouse.getY()/GUI.scale)
         local swipe_length = self.finger_down - y
         self.movingy = self.movingy - swipe_length/7
         self.finger_down = nil
      else
         local x= math.max(self.x,love.mouse.getX()/GUI.scale)
         local swipe_length = self.finger_down - x
         self.movingx = self.movingx - swipe_length/7
         self.finger_down = nil
      end
   end
   return self
end


function TScrollBar:onDraw()
   if not TControl.onDraw(self) then return end


   if self.orientation == "vertical" then
      lg.setColor(self.bgColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("fill", self.x, self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
      lg.setColor(self.fontColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("line", self.x, self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
      lg.rectangle("line", self.x + self.w/2, self.y+self.w/2, 1, self.h-self.w, 8) -- the very line/path
      lg.setColor(self.bgcol or self.bgColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("fill", self.x+self.w/4, self.y + (self.knobpos or 0)+self.w/4, self.w/2, self.w/2,8) -- knob fill
      lg.setColor(self.fontColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("line", self.x+self.w/4, self.y + (self.knobpos or 0)+self.w/4, self.w/2, self.w/2,8) -- knob fill
   else
      lg.setColor(self.bgColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("fill", self.x, self.y, self.w, self.h, 8)
      lg.setColor(self.fontColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("line", self.x, self.y, self.w, self.h, 8)
      lg.rectangle("line", self.x + self.h/2, self.y+self.h/2, self.w-self.h, 1,8) -- the very line/path
      lg.setColor(self.bgcol or self.bgColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("fill", self.x + (self.knobpos or 0)+self.h/4, self.y+self.h/4, self.h/2, self.h/2,8) -- knob fill
      lg.setColor(self.fontColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("line", self.x + (self.knobpos or 0)+self.h/4, self.y+self.h/4, self.h/2, self.h/2,8) -- knob outline
   end
end

--------------------------------------------------------------



----- TPanel -------------------------------------------------
local TPanel = {ancestor = "TControl", type = "TPanel"}
TPanel.__index = TPanel

setmetatable(TPanel, {
  __index = TControl, -- this is what makes the inheritance work
  __call = function (cls, ...)
    --print(...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})

function TPanel:_init(...)
  TControl._init(self, ...) -- call the base class constructor
  self.type = "TPanel"
  self.ancestor = "TControl"
  self.borders = true
  self.curved_borders = 0
end

function TPanel:setText(caption, align, valign)
   align = align or self.text_align
   self.text_align = align

   valign = valign or self.text_vertical_align
   self.text_vertical_align = valign

   self.caption = caption
   local font = lg.getFont()
   local _, text   = font:getWrap(caption, self.w-20)
   self.text_length = utf8.len(table.concat(text,"\n"))
   self.text = text
   if self.autosize then
      local wid = 0
      for row, line in ipairs(self.text) do
          if wid < font:getWidth(line) then
             wid = font:getWidth(line)
          end
      end
      self.w = wid+60

   end
   return self
end

function TPanel:getText()
   return self.caption or ""
end

function TPanel:drawOutline(text,x,y)
   local col = {lg.getColor()}
   lg.setColor(100,100,100,50)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   for i=-2,2 do
       for j=-2,2 do
           lg.print(text,i+x,j+y)
       end
   end
   lg.setColor(col)
end

function TPanel:setTextOutline(param)
   self.outline = param
   return self
end

function TPanel:toggleDarkenBG(param)
   self.darkenBG = param
   return self
end

function TPanel:onDraw()
   if not TControl.onDraw(self) then return end
   --lg.setStencilTest()
   local col = {lg.getColor()}
   if self.darkenBG then
      lg.setColor(55,55,105,180)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("fill", 0, 0, GUI.width, GUI.height)
   end

   lg.setColor(self.bgcol or self.bgColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("fill",self.x,self.y,self.w,self.h, self.rounded_corners or GUI.rounded_corners)
   lg.setColor(self.fontColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   if self.borders then
      lg.rectangle("line",self.x,self.y,self.w,self.h, self.rounded_corners or GUI.rounded_corners)
   end
   lg.setColor(col)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   if self.caption then
      lg.setColor(self.fontColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      local font = lg.getFont()
      local lineheight = font:getHeight("I")
      local total_height = 0
      for row, line in ipairs(self.text) do
          total_height = total_height + font:getHeight(line)
      end

      local y_offset = 0
      if     self.text_vertical_align == "top"    then
             y_offset = 0 -- let it sit here, m'kay?
      elseif self.text_vertical_align == "center" then
             y_offset = math.ceil((self.h-total_height)/2)
      elseif self.text_vertical_align == "bottom" then
             y_offset = self.h-lineheight - 0
      end

      for row, line in ipairs(self.text) do
          local linewidth = font:getWidth(line)
          if self.text_align == "left" then
             if self.outline then self:drawOutline(line, self.x+10, self.y+(row-1)*(lineheight)+y_offset)  end
             lg.print(line, self.x+10, self.y+(row-1)*(lineheight)+y_offset)
          elseif self.text_align == "right" then
            if self.outline then self:drawOutline(line, self.x+self.w-10-linewidth, self.y+(row-1)*(lineheight)+y_offset)  end
            lg.print(line, self.x+self.w-10-linewidth, self.y+(row-1)*(lineheight)+y_offset)
          elseif self.text_align == "center" then
            if self.outline then self:drawOutline(line, self.x+math.ceil((self.w-linewidth)/2), self.y+(row-1)*(lineheight)+y_offset)  end
            lg.print(line, self.x+math.ceil((self.w-linewidth)/2), self.y+(row-1)*(lineheight)+y_offset)
          end
      end
   end
end
----------------------------------------------------------



----------------------------------------------------------
----- TGrid ----------------------------------------------
local TGrid = {ancestor = "TControl", type = "TGrid"}
TGrid.__index = TGrid

setmetatable(TGrid, {
  __index = TControl, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})

function TGrid:_init(...)
  TControl._init(self, ...) -- call the base class constructor
  self.type = "TGrid"
  self.ancestor = "TControl"
  self.borders = true
  self.cells={}
  self.widths = {}
  self.heights = {}
  self:setCellSize(0,0)
  self:setSize(0,0)
end

function TGrid:setCellSize(x,y)
   --print("cell size",x,y)
   self.cellsize = {x,y}
   return self
end

function TGrid:setSize(x,y)
   local w, h = 0, 0
   for col=1, x do
       self.cells[col] = self.cells[col] or {}
       h = 0
       for row = 1, y do
           --print(col,row,self.cellsize, self.cells[col])
           local data = self.cells[col][row] or {w = self.cellsize[1] or 0, h=self.cellsize[2] or 0, text = "", icons = {}}
           self.cells[col][row] = data
           if data.h >= (self.heights[row] or 0) then
              self.heights[row] = data.h
           end
           if data.w >= (self.widths[col] or 0) then
              self.widths[col] = data.w
           end
           h = h + (self.heights[row] or 0)
       end
       w = w + (self.widths[col] or 0)
   end
   self.cols = math.max(x, self.cols or 0)
   self.rows = math.max(y, self.rows or 0)
   self.w  = w --- 1 -- #hotfix #3
   self.h  = h --- 1 -- #hotfix #2
   return self
end

function TGrid:getSize()
   self.cols = self.cols or 0
   self.rows = self.rows or 0
   return self.cols, self.rows
end

function TGrid:getRowHeight(row)
   return self.heights[row] or 0
end

function TGrid:getColWidth(col)
   return self.widths[col] or 0
end

function TGrid:getCellData(x,y)
   return self.cells[x][y]
end

function TGrid:setCellData(x,y,text, size, icons, outlineColor, textColor, bgColor, object, dbg)
   
   --if dbg then print(x,y,text, size, icons, table.concat(outlineColor or {},","), table.concat(textColor or {},","), table.concat(bgColor or {},","), dbg) end
   local cols, rows = self:getSize()
   if cols<x or rows<y then self:setSize(x,y) end

   local olddata = self.cells[x][y]
   local size = size or {olddata.w, olddata.h} or self.cellsize -- cellsize is a TABLE
   --print("size["..x.."]["..y.."]",size[1],size[2])
   local data = {--text   or olddata.text,
                 icons=icons or olddata.icons or {},
                 w = size[1] or 0,
                 h = size[2] or 0,
                 outlineColor=outlineColor or olddata.outlineColor,
                 fontColor=textColor or olddata.fontColor,
                 bgColor=bgColor or olddata.bgColor or self.bgColor or GUI.bgColor,
                 bgcol=olddata.bgcol,
                 object = object,
                 text = text,
                 caption = olddata.caption,
              }
   data.borders = true--(data.outlineColor~=nil)
   --GUI.debug_panel:setText("Cell "..x..":"..y.."; bgcol="..table.concat(olddata.bgcol or {},",").."; bgColor"..table.concat(olddata.bgColor or {},",").. "; "..tostring(dbg))
   self.cells[x][y] = data
   if c or olddata.text then
       self:_setcelltext(x,y, text or olddata.text, align or olddata.align, valign or olddata.valign)
       if (not olddata.bgcol) and data.bgColor then
          self.cells[x][y].bgcol = data.bgColor
       end

       local txt = data.caption       
       for i=1, #txt do
           data.w = math.max(math.max(10+lg.getFont():getWidth(txt[i] or ""), data.w), (self.widths[x] or 0) )
       end

       self:_setcelltext(x,y, text or olddata.text, align or olddata.align, valign or olddata.valign)

       if (self.heights[y] or 0) < data.h or (self.widths[x] or 0) < data.w then
          self:setSize(self.cols, self.rows)
       end
   end
  if dbg then print(self.cells[x][y].object) end
   return self
end

function TGrid:_setcelltext(x,y, text, align, valign)
   --print(x,y, text, align, valign)
   local grid = self
   local self = grid.cells[x][y]

   align = align or "center"
   self.align = align

   valign = valign or "center"
   self.valign = valign

   self.caption = caption or ""
   local font = lg.getFont()
   local _, _text   = font:getWrap(text, self.w-2)
   self.text_length = utf8.len(table.concat(_text,"\n"))
   self.caption = _text
   self.text = text
   if self.autosize then
      local wid = 0
      for row, line in ipairs(self.caption) do
          if wid < font:getWidth(line) then
             wid = font:getWidth(line)
          end
      end
      self.w = wid+60
   end
   return grid
end

function TGrid:_drawcell(x,y)
   local col  = {lg.getColor()}
   local grid = self
   local self = grid.cells[x][y]
   --print("cell", x,y,"--------------------------------------------------")
   --for k,v in pairs(self) do print(k,v) end

   lg.setColor(self.bgColor or self.bgcol or grid.bgColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   --lg.setLineWidth(1)
   --self.x,self.y,self.w,self.h = math.ceil(self.x), math.ceil(self.y), math.ceil(self.w), math.ceil(self.h)
   lg.rectangle("fill",self.x,self.y,self.w,self.h, self.rounded_corners or GUI.rounded_corners)

   lg.setColor(self.outlineColor or grid.outlineColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end

   if self.borders then
      lg.rectangle("line",self.x,self.y,self.w,self.h, self.rounded_corners or GUI.rounded_corners)
   end

   lg.setColor(col)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end

   if self.caption then
      lg.setColor(self.fontColor or grid.fontColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      local font = lg.getFont()
      local lineheight = font:getHeight("I")
      local total_height = 0

      for row, line in ipairs(self.caption) do
          total_height = total_height + font:getHeight(line)
      end

      local y_offset = 0
      if     self.valign == "top"    then
             y_offset = 2 -- let it sit here, m'kay?
      elseif self.valign == "center" then
             y_offset = math.ceil((self.h-total_height)/2)
      elseif self.valign == "bottom" then
             y_offset = self.h-lineheight - 2
      end

      for row, line in ipairs(self.caption) do
          local linewidth = font:getWidth(line)
          if self.align == "left" then
             if self.outline then self:drawOutline(line, self.x+10, self.y+(row-1)*(lineheight)+y_offset)  end
             lg.print(line, self.x+10, self.y+(row-1)*(lineheight)+y_offset)
          elseif self.align == "right" then
            if self.outline then self:drawOutline(line, self.x+self.w-linewidth, self.y+(row-1)*(lineheight)+y_offset)  end
            lg.print(line, self.x+self.w-10-linewidth, self.y+(row-1)*(lineheight)+y_offset)
          elseif self.align == "center" then
            if self.outline then self:drawOutline(line, self.x+math.ceil((self.w-linewidth+4)/2), self.y+(row-1)*(lineheight)+y_offset)  end
            lg.print(line, self.x+math.ceil((self.w-linewidth+4)/2), self.y+(row-1)*(lineheight)+y_offset)
          end
      end
   end

   return grid
end

function TGrid:onDraw()
   lg.rectangle("line", self.x, self.y, self.w, self.h)
   if not TControl.onDraw(self) then return end

   local col = {lg.getColor()}
   if self.darkenBG then
      lg.setColor(55,55,55,180)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      lg.rectangle("fill", 0, 0, 2000, 2000)
   end

   local w = 0
   for col = 1, self.cols do
       local h = 0
       local wid = self.widths[col]
       for row = 1, self.rows do
           local data = self.cells[col][row]
           local hei = self.heights[row]
           data.x = self.x+w
           data.y = self.y+h
           data.w = wid
           data.h = hei
           self:_drawcell(col,row)

--[[
           lg.setColor(data.bgColor or self.bgColor)
           if self.transparency then
              local col = {lg.getColor()}
              col[4] = self.transparency
              lg.setColor(col)
           end
           lg.rectangle("fill",self.x+w,self.y+h, wid, hei)

           lg.setColor(data.outlineColor or self.outlineColor)
           if self.transparency then
              local col = {lg.getColor()}
              col[4] = self.transparency
              lg.setColor(col)
           end

           lg.rectangle("line",self.x+w,self.y+h, wid, hei)

           lg.setColor(data.fontColor or self.fontColor)
           local font = lg.getFont()
           local _w = font:getWidth(data.text)
           local _h = font:getHeight(data.text)
           lg.print(data.text, self.x+w+math.floor((wid-_w)/2),self.y+h+math.floor((hei-_h)/2))
]]
           h = h+hei
       end
       w = w+wid
   end
   return self
end

function TGrid:getCellAt(x,y)
   local w = self.x
   local h = self.y
   local retx, rety = 0,0
   for i=1, self.cols do
       w = w+self.widths[i]
       if w > x then
          retx = i
          break
       end
   end
   for i=1, self.rows do
       h = h+self.heights[i]
       if h > y then
          rety = i
          break
       end
   end
   return retx, rety
end

function TGrid:onHover(x,y)
   if self.hightlightOnHover then
      local x,y = x/(GUI.scalex or GUI.scale), y/(GUI.scaley or GUI.scale)
       local _x,_y = self:getCellAt(x,y)
       for i = 1, self.cols do
           for j=1, self.rows do
               if i==_x and j==_y then
                  local data = self:getCellData(i,j)
                  self:setCellData(_x,_y, nil, nil, nil, nil, nil, {150,150,200})
               else
                  local data = self:getCellData(i,j)
                  self:setCellData(i,j, nil, nil, nil, nil, nil, data.bgcol)
               end
           end
       end
    end
end

function TGrid:onMouseLeave()
   if self.hightlightOnHover then
      for i = 1, self.cols do
          for j=1, self.rows do
              local data = self:getCellData(i,j) or {}
              self:setCellData(i,j, nil, nil, nil, nil, nil, data.bgcol)
          end
      end
   end
end
----------------------------------------------------------

--------------- TButton ----------------------------------

local TButton = {ancestor = "TPanel", type="TButton"}
TButton.__index = TButton

setmetatable(TButton, {
  __index = TPanel, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})

function TButton:_init(...)
  TPanel._init(self, ...) -- call the base class constructor
  self.type = "TButton"
  self.ancestor = "TPanel"
end

function TButton:update()
   if self.down then
      local color = {self.bgColor[1],self.bgColor[2],self.bgColor[3],self.bgColor[4]}
      color[1],color[2],color[3],color[4] = color[1]-20,color[2]-20,color[3]-20, (255)
      self.bgcol = color
      return self
   end
end

function TButton:onHover()
   local color = {self.bgColor[1],self.bgColor[2],self.bgColor[3],self.bgColor[4]}
   color[1],color[2],color[3],color[4] = color[1]+10,color[2]+10,color[3]+50, (color[4] or 255)+20
   self.bgcol = color
   return self
end

function TButton:onMouseLeave()
   self.bgcol = nil
   return self
end

----------------------------------------------------------

---- TLabel -----------------------------------------------------

local TLabel = {ancestor = "TControl", type = "TLabel"}
TLabel.__index = TLabel

setmetatable(TLabel, {
  __index = TControl, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})

function TLabel:_init(...)
  TControl._init(self, ...) -- call the base class constructor
  self.type = "TLabel"
  self.bgcol = {0,0,0,0}
  self.ancestor = "TControl"
end

function TLabel:setText(caption, align, valign)
   align = align or self.text_align
   self.text_align = align

   valign = valign or self.text_vertical_align
   self.text_vertical_align = valign

   self.caption = caption
   local font = lg.getFont()
   local _, text   = font:getWrap(caption, self.w-20)
   self.text_length = utf8.len(table.concat(text,"\n"))
   self.text = text
   if self.autosize then
      local font = lg.getFont()
      local lineheight = font:getHeight("I")
      local total_height = 0
      for row, line in ipairs(self.text) do
          total_height = total_height + font:getHeight(line)
      end
      self.h = total_height + 4
   end
   return self
end

function TLabel:getText()
   return self.caption, self.text_align, self.text_vertical_align
end


function TLabel:drawOutline(text,x,y)
   local col = {lg.getColor()}
   lg.setColor(0,0,0,50)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   for i=-2,2 do
       for j=-2,2 do
           lg.print(text,i+x,j+y)
       end
   end
   lg.setColor(col)
end

function TLabel:setTextOutline(param)
   self.outline = param
   return self
end

function TLabel:onDraw()
   if not TControl.onDraw(self) then return end

   lg.setColor(self.bgcol or self.bgColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("fill",self.x,self.y,self.w,self.h, self.rounded_corners or GUI.rounded_corners)

   if self.caption then
      lg.setColor(self.fontColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      local font = lg.getFont()
      local lineheight = font:getHeight("I")
      local total_height = 0
      for row, line in ipairs(self.text) do
          total_height = total_height + font:getHeight(line)
      end

      local y_offset = 0
      if     self.text_vertical_align == "top"    then
             y_offset = 10 -- let it sit here, m'kay?
      elseif self.text_vertical_align == "center" then
             y_offset = math.ceil((self.h-lineheight-10)/2)
      elseif self.text_vertical_align == "bottom" then
             y_offset = self.h-lineheight - 10
      end

      for row, line in ipairs(self.text) do
          local linewidth = font:getWidth(line)
          if self.text_align == "left" then
             if self.outline then self:drawOutline(line, self.x+10, self.y+(row-1)*(lineheight)+y_offset)  end
             lg.print(line, self.x+10, self.y+(row-1)*(lineheight)+y_offset)
          elseif self.text_align == "right" then
            if self.outline then self:drawOutline(line, self.x+self.w-10-linewidth, self.y+(row-1)*(lineheight)+y_offset)  end
            lg.print(line, self.x+self.w-10-linewidth, self.y+(row-1)*(lineheight)+y_offset)
          elseif self.text_align == "center" then
            if self.outline then self:drawOutline(line, self.x+math.ceil((self.w-linewidth)/2), self.y+(row-1)*(lineheight)+y_offset)  end
            lg.print(line, self.x+math.ceil((self.w-linewidth)/2), self.y+(row-1)*(lineheight)+y_offset)
          end
      end
   end
end


-----------------------------------------------------------------

------------------------------------------------------------------------------------

----- TStatsButton -----------------------------------------------------------------
local TStatsButton = {ancestor = "TButton", type="TStatsButton"}
TStatsButton.__index = TStatsButton

setmetatable(TStatsButton, {
  __index = TButton, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})

function TStatsButton:_init(...)
  TButton._init(self, ...) -- call the base class constructor
  self.type = "TStatsButton"
  self.ancestor = "TButton"

end

--function TStatsButton:onHover()
--end

function TStatsButton:onHover()
   local color = {self.bgColor[1],self.bgColor[2],self.bgColor[3],self.bgColor[4]}
   color[1],color[2],color[3],color[4] = color[1]+100,color[2]+50,color[3]+100, (color[4] or 255)+20
   self.bgcol = color
   return self
end

function TStatsButton:onMouseLeave()
   self.bgcol = nil
   return self
end

function TStatsButton:setData(data)
   self.data = data
   if self.datatype=="class" then
      class = {class=classname, growth=growth, basehp=hp, basemp=mp, baseatk=atk, basedef=def, baseagi=agi, basemov=mov, sprite=sprite, tier = tier}
      data.hp, data.maxhp = data.basehp, data.basehp
      data.mp, data.maxmp = data.basemp, data.basemp
      data.name = "1234567890"

   elseif self.datatype=="instance" then
   end
   local x, y = self.x, self.y
   local font = lg.getFont()
   local hp,maxhp = acronum(data.hp), acronum(data.maxhp)
   local mp,maxmp = acronum(data.mp), acronum(data.maxmp)
   local hptext = hp .."/" ..maxhp
   local mptext = mp .."/" ..maxmp
   local classname = (data.class or "") .."\t"..(data.level or "L 1")
   local hpwidth = font:getWidth("HP")
   local mpwidth = font:getWidth("MP")
   local hptextwidth = font:getWidth(hptext)
   local mptextwidth = font:getWidth(mptext)
   local classnamewidth = font:getWidth(classname)
   local namewidth = font:getWidth(data.name)
   local h = 75
   local w = math.max(hpwidth+hptextwidth,mpwidth+mptextwidth,namewidth)+55+classnamewidth
   self.h = h -- а это - правильно
   self.w = w

   return self
end

function TStatsButton:switchDataType(datatype)  -- class or instance
   self.datatype = datatype
   return self
end

function TStatsButton:onDraw()
   --if not TButton.onDraw(self) then return end

   local data = self.data
   local data = self.data
   local x, y = self.x, self.y
   local font = lg.getFont()
   local hp,maxhp = acronum(data.hp), acronum(data.maxhp)
   local mp,maxmp = acronum(data.mp), acronum(data.maxmp)
   local hptext = hp .."/" ..maxhp
   local mptext = mp .."/" ..maxmp
   local classname = (data.class or "") .."\t"..(data.level or "L 1")
   local hpwidth = font:getWidth("HP")
   local mpwidth = font:getWidth("MP")
   local hptextwidth = font:getWidth(hptext)
   local mptextwidth = font:getWidth(mptext)
   local classnamewidth = font:getWidth(classname)
   local namewidth = font:getWidth(data.name)
   local h = self.h
   local w = self.w
   local hpbarleft  = x+hpwidth+20
   local mpbarleft  = x+mpwidth+20
   local hpbarwidth = w-hpwidth-hptextwidth-40
   local mpbarwidth = w-mpwidth-mptextwidth-40
   --if data.maxhp <100 then
      hpbarleft = hpbarleft+hpbarwidth * (100-data.maxhp)/100
      hpbarwidth = hpbarwidth * data.maxhp/100
   --end
   --if data.maxhp <100 then
      mpbarleft = mpbarleft+mpbarwidth * (100-data.maxmp)/100
      mpbarwidth = mpbarwidth * data.maxmp/100
   --end
   local name = (data.name or "Unknown")
   local classnamex = w-classnamewidth-10
   local hptextx = w-hptextwidth-10
   local mptextx = w-mptextwidth-10
   local hppercentage = math.min(math.ceil(100*data.hp/data.maxhp),100)
   local mppercentage = math.min(math.ceil(100*data.mp/data.maxmp),100)

   --local function hpbarstencil() lg.rectangle("fill", mpbarleft, y+35, hppercentage, 6) end
   --local function mpbarstencil() lg.rectangle("fill", mpbarleft, y+55, mppercentage, 6) end
   --local comparemode, comparevalue = lg.getStencilTest()
   --local last_stencil
   --if lg.last_stencil then
--    last_stencil = {lg.last_stencil[1],lg.last_stencil[2],lg.last_stencil[3],lg.last_stencil[4]}
-- end

   --print("width",w, hptextx )

   lg.setColor(self.bgcol or self.bgColor)
   lg.rectangle("fill", x, y, w, h,4)
   lg.setColor(self.fontColor)
   lg.rectangle("line", x, y, w, h,4)
   lg.print(name, x+10,y-12)
   lg.print(classname, x+classnamex ,y-12)


   lg.print("HP", x+10,y+12)
   lg.print(hptext, x+hptextx,y+12)
   black()
   lg.rectangle("fill", hpbarleft, y+35, hpbarwidth, 10,4)
   degrees = hppercentage*0.8;
   rotated()
   lg.rectangle("fill", hpbarleft, y+35, hppercentage*hpbarwidth/100, 10,4)
   lg.setColor(self.fontColor)
   lg.rectangle("line", hpbarleft, y+35, hpbarwidth, 10,4)


   lg.print("MP", x+10,y+34)
   lg.print(mptext, x+mptextx,y+34)
   black()
   lg.rectangle("fill", mpbarleft, y+58, mpbarwidth, 10,4)
   degrees = mppercentage*0.8;
   rotated()
   lg.rectangle("fill", mpbarleft, y+58, mppercentage*mpbarwidth/100, 10,4)
   lg.setColor(self.fontColor)
   lg.rectangle("line", mpbarleft, y+58, mpbarwidth, 10,4)
   return self
end
------------------------------------------------------------------------------------


local TImage = {type = "TImage", anccestor = "TControl"}

TImage.__index = TImage

setmetatable(TImage, {
  __index = TControl, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})


function TImage:_init(...)
  TControl._init(self, ...) -- call the base class constructor
  self.type = "TImage"
  self.ancestor = "TControl"
end

function TImage:loadFile(filename)
   self.image = lg.newImage(filename)
   self.imagewidth = self.image:getWidth()
   self.imageheight = self.image:getHeight()
   if self.tile then
      self.repeatx = math.ceil(self.w/self.imagewidth)*2
      self.repeaty = math.ceil(self.h/self.imageheight)*2
   end
   return self
end

function TImage:setTile(flag)
   self.tile = flag
   return self
end

function TImage:startScroll(x,y)
   self.scrollx = x
   self.scrolly = y
   return self
end

function TImage:stopScroll()
   self.scrollx = nil
   self.scrolly = nil
   return self
end


function TImage:update(dt)
   --print(self.image, self.tile, self.scrollx, self.scrolly)
   if self.image and self.tile and (self.scrollx or self.scrolly) then
      self.x = self.x + dt*(self.scrollx or 0)
      self.y = self.y + dt*(self.scrolly or 0)
      --print(self.x,self.imagewidth)
      if -self.x>self.imagewidth  then self.x = self.x + self.imagewidth  end
      if -self.y>self.imageheight then self.y = self.y + self.imageheight end

      if -self.x<0 then self.x = - self.imagewidth  end
      if -self.y<0 then self.y = - self.imageheight end

      self.w = self.parent.w
      self.h = self.parent.h

      self.repeatx = math.ceil(self.w/self.imagewidth)*2
      self.repeaty = math.ceil(self.h/self.imageheight)*2
   end
end

function TImage:onDraw(filename)
   if not TControl.onDraw(self) then return end
   if self.tile and self.repeatx and self.repeaty then
      for i=1, self.repeatx do
          for j=1, self.repeaty do
              lg.draw(self.image, self.x+(i-1)*self.imagewidth, self.y+(j-1)*self.imageheight)
          end
      end
   else
      lg.draw(self.image, self.x, self.y,0, self.w/self.imagewidth , self.h/self.imageheight)
   end
   return self
end
------------------------------------------------------------------------------------


local TEdit = {type = "TEdit", anccestor = "TControl"}

TEdit.__index = TEdit

setmetatable(TEdit, {
  __index = TControl, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})


function TEdit:_init(...)
  TControl._init(self, ...) -- call the base class constructor
  self.type = "TEdit"
  self.ancestor = "TControl"
  self.bgColor = {025,025,055, 200}
  self.h = 40
  self.fontColor = {255,255,255}
end

function TEdit:setText(text)
   self.text = text
   if text=='' then
      GUI:resetTextBuffer() 
   end
   return self
end

function TEdit:onDraw()
   if not TControl.onDraw(self) then return end
   lg.setColor(self.bgColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("fill", self.x,self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
   lg.setColor(self.fontColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("line", self.x,self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
   local h = lg.getFont():getHeight("self.text")
   lg.print(self.text or "",self.x+5,self.y+math.ceil((self.h-h)/2))
   if GUI.focused_input == self then
      local wid = lg.getFont():getWidth(self.text or "")
      lg.setColor(255,255,255, 255*math.abs(math.sin(GUI.sin_timer)))
      
      lg.print("_",wid+self.x+8,self.y+math.ceil((self.h-h)/2))
      lg.setColor(255,255,255)
   end
   return self
end

function TEdit:onClick()
   GUI.focused_input = self
   GUI.text_buff = (GUI.focused_input.text or "")
   if love.system.getOS()=="Android" then love.keyboard.setTextInput(true) end
end







----- TProgressBar -------------------------------------------
local TProgressBar = {ancestor = "TPanel", type = "TProgressBar"}
TProgressBar.__index = TProgressBar

setmetatable(TProgressBar, {
  __index = TPanel, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    self.progress = 0
    self.h = lg.getFont():getHeight(self.progress or "%")
    return self
  end,
})

function TProgressBar:_init(...)
  TPanel._init(self, ...) -- call the base class constructor
  self.type = "TProgressBar"
  self.ancestor = "TPanel"
end

function TProgressBar:setProgress(num)
   if not num then return self end
   num = round(num,2)
   self.progress = num
   self:setText(num.."%", "center", "center")
   return self
end

function TProgressBar:onDraw()
   if self.progress >0 then
      local bgColor = {rotate_hue(clred, self.progress*0.8, -self.progress*0.8)}
      self.bgColor[4] = 120
      lg.setColor(bgColor)
      lg.rectangle("fill",self.x, self.y, self.progress*self.w/100, self.h, self.rounded_corners or GUI.rounded_corners)
   end
   if not TPanel.onDraw(self) then return end
end


-----------------------------------------------------------------

local TCheckBox = {ancestor = "TButton", type="TCheckBox"}
TCheckBox.__index = TCheckBox

setmetatable(TCheckBox, {
  __index = TButton, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})

function TCheckBox:_init(...)
  TButton._init(self, ...) -- call the base class constructor
  self.type = "TCheckBox"
  self.ancestor = "TButton"
  self.autosize = true
end


function TCheckBox:onHover()
   local color = {self.bgColor[1],self.bgColor[2],self.bgColor[3],self.bgColor[4]}
   color[1],color[2],color[3],color[4] = color[1]+100,color[2]+50,color[3]+100, (color[4] or 255)+20
   self.bgcol = color
   return self
end

function TCheckBox:onMouseLeave()
   self.bgcol = nil
   return self
end

function TCheckBox:onClick()
   self.checked = not not not self.checked
   if self.checked then
      local color = {self.bgColor[1],self.bgColor[2],self.bgColor[3],self.bgColor[4]}
      color[1],color[2],color[3],color[4] = color[1]+200,color[2]+150,color[3]+200, (color[4] or 255)+20
      self.bg2col = color
   else
      local color = {self.bgColor[1],self.bgColor[2],self.bgColor[3],self.bgColor[4]}
      color[1],color[2],color[3],color[4] = color[1], color[2], color[3], (color[4] or 255)+20
      self.bg2col = color
   end
   return self
end

--[[
function TCheckBox:update(dt)
   if not self.timer___ then
      self.h = math.random(200)
      self.timer___ = 2
   else
      self.timer___ = self.timer___ - dt
      if self.timer___<0 then self.timer___ = nil end
   end
   return self
end ]]

function TCheckBox:onDraw()
   if not TControl.onDraw(self) then return end


   local col = {lg.getColor()}
--[[
   -- outline filling
   lg.setColor(self.bgcol or self.bgColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("fill",self.x,self.y,self.w,self.h,8)
   lg.setColor(self.fontColor)

   -- outline outline
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("line",self.x,self.y,self.w,self.h,8)
   lg.setColor(col)
]]

   -- box filling
   lg.setColor(self.bg2col or self.bgColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("fill",self.x,self.y+(self.h-40)/2,40,40, self.rounded_corners or GUI.rounded_corners)
   lg.setColor(self.fontColor)

   -- box outline
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("line",self.x, self.y+(self.h-40)/2,40,40, self.rounded_corners or GUI.rounded_corners)
   if self.checked then
      local font = lg.getFont()      
      lg.print("+",self.x+math.floor((60-font:getWidth("+"))/2), self.y+ math.floor((self.h-font:getHeight("+")*2)/2), math.rad(45), 2)
   end
   lg.setColor(col)


   -- caption
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   if self.caption then
      lg.setColor(self.fontColor)
      if self.transparency then
         local col = {lg.getColor()}
         col[4] = self.transparency
         lg.setColor(col)
      end
      local font = lg.getFont()
      local lineheight = font:getHeight("I")
      local total_height = 0
      for row, line in ipairs(self.text) do
          total_height = total_height + font:getHeight(line)
      end

      local y_offset = 0
      y_offset = math.ceil((self.h-total_height)/2)

      for row, line in ipairs(self.text) do
          local linewidth = font:getWidth(line)
          if self.text_align == "left" then -- сдвигаем только если выравнивание по левому краю. пока.
             if self.outline then self:drawOutline(line, 40+ self.x+10, self.y+(row-1)*(lineheight)+y_offset)  end
             lg.print(line, 40+ self.x+10, self.y+(row-1)*(lineheight)+y_offset)
          elseif self.text_align == "right" then
            if self.outline then self:drawOutline(line, self.x+self.w-10-linewidth, self.y+(row-1)*(lineheight)+y_offset)  end
            lg.print(line, self.x+self.w-10-linewidth, self.y+(row-1)*(lineheight)+y_offset)
          elseif self.text_align == "center" then
            if self.outline then self:drawOutline(line, self.x+math.ceil((self.w-linewidth)/2), self.y+(row-1)*(lineheight)+y_offset)  end
            lg.print(line, self.x+math.ceil((self.w-linewidth)/2), self.y+(row-1)*(lineheight)+y_offset)
          end
      end
   end
end

------------------------------------------------------------------------------------


local TPasswordEdit = {type = "TPasswordEdit", anccestor = "TControl"}

TPasswordEdit.__index = TPasswordEdit

setmetatable(TPasswordEdit, {
  __index = TEdit, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})


function TPasswordEdit:_init(...)
  TControl._init(self, ...) -- call the base class constructor
  self.type = "TPasswordEdit"
  self.ancestor = "TEdit"
  self.bgColor = {025,025,055, 200}
  self.fontColor = {255,255,255}
  self.h = 40
end

function TPasswordEdit:onDraw()
   if not TControl.onDraw(self) then return end
   lg.setColor(self.bgColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("fill", self.x,self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
   lg.setColor(self.fontColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("line", self.x,self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
   lg.print(self.text or "",self.x+5,self.y-5)
   if GUI.focused_input == self then
      local wid = lg.getFont():getWidth(self.text or "")
      local hei = lg.getFont():getHeight("M")
      lg.setColor(255,255,255, 255*math.abs(math.sin(GUI.sin_timer)))
      lg.line(wid+self.x+8,self.y+self.h-5, wid+self.x+20, self.y+self.h-5)
      lg.setColor(255,255,255)
   end
   return self
end

function TPasswordEdit:setText(text)
   self.real_text = text
   self.text = string.rep("*",#(text or ""))
   return self
end

function TPasswordEdit:getText()
   return self.real_text
end

------------------------------------------------------------------------------------

local TComboBox = {type = "TComboBox", ancestor = "TEdit"}

TComboBox.__index = TComboBox

setmetatable(TComboBox, {
  __index = TEdit, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})


function TComboBox:_init(x,y,w,h, init_data)
  TEdit._init(self, x,y,w,h, init_data) -- call the base class constructor
  self.type = "TComboBox"
  self.ancestor = "TControl"
  --self.bgColor = GUI.bgColor or {025,025,055, 200}
  --self.h = {...}--x,y,w,h, init_data
  self.h = h
  --self.fontColor = {255,255,255}
  self.max_dropped_items = (init_data or {}).max_dropped_items or 3
  self.items = {}
  local itempanel_h = self.h*math.min(self.max_dropped_items, math.max(#self.items, 1))
  local itempanel_y = self.y+self.h
  if itempanel_h + itempanel_y > GUI.height then
     itempanel_y = self.y - itempanel_h
  end
  self.itempanel = GUI.newScrollBox(self.x,itempanel_y,self.w,itempanel_h):setTransparency(255)
  self.itempanel.fill = true
  self.itempanel.topmost = true
  
  --self.debug_bounds = true
  --self.itempanel.debug_bounds = true
  self.dropdownbutton = GUI.newButton(self.w-self.h,0,self.h,self.h):setParent(self):setText("v", "center","center"):show()
  self.dropdownbutton.onClick = function()
     if self.itempanel.visible then
        self.dropdownbutton:setText("v", "center","center")
        self.itempanel:hide()
     else
        self.dropdownbutton:setText("^", "center","center")
        self.itempanel:show()
     end
  end

  self.onClick = function()
     if self.itempanel.visible then
        self.dropdownbutton:setText("v", "center","center")
        self.itempanel:hide()
     end
  end

  return self
end

function TComboBox:setItems(items)
   for i, child in ipairs(self.children) do -- iterate children only
       if self.children[i].is_item then
          self.children[i] = nil
       end
   end
   local itempanel_h = self.h*math.min(self.max_dropped_items, math.max(#items, 1))
   local itempanel_y = self.y+self.h
   if itempanel_h + itempanel_y > GUI.height then
      itempanel_y = self.y - itempanel_h
   end

   self.itempanel:setPosition(nil, itempanel_y)
   self.itempanel:setSize(nil, itempanel_h)
   self.itempanel.topmost = true
   local _tmp = self
   for i, string in ipairs(items) do
       local item = GUI.newButton(0,(i-1)*self.h,self.w-40,self.h):setText(string,"left","center"):setParent(_tmp.itempanel):show()
       --item.internal_id = i
       function item:onClick()
          local text = item:getText()
          _tmp.itempanel.topmost = true
          _tmp.itempanel:hide()
          _tmp.dropdownbutton:setText("v", "center","center")
          _tmp:setText(text)
          _tmp.selected_item = i          
          if _tmp.onSelect then
             _tmp:onSelect(_tmp:selectedItem())
          end
       end
       item.is_item = true
       item.borders = false
       self.items[#self.items+1] = item
   end

   return self
end
--[[
function TComboBox:onSelect(item)
   self:setText(item.caption)
   self.selected_item = item.internal_id
end]]
function TComboBox:setIndex(idx)   
   self:setText(self.items[idx]:getText())
   self.selected_item = idx
   return self
end

function TComboBox:getIndex()
   return self.selected_item
end

function TComboBox:selectedItem()
   return self.text, self.selected_item
end

function TComboBox:setParent(control)
   TEdit.setParent(self, control)
   self.itempanel:setParent(control):hide()
   return self
end

function TComboBox:setText(text)
   self.text = text
   return self
end

function TComboBox:onDraw()
   if not TControl.onDraw(self) then return end
   lg.setColor(self.bgColor)   
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("fill", self.x,self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
   lg.setColor(self.outlineColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("line", self.x,self.y, self.w, self.h, self.rounded_corners or GUI.rounded_corners)
   local t = self.text or ""
   local f = lg.getFont()
   local w,h = f:getWidth(t), f:getHeight(t)
   lg.print(t, self.x+math.floor((self.w-self.h-w)/2),self.y+math.floor((self.h-h)/2))
   if GUI.focused_input == self then
      local wid = lg.getFont():getWidth(self.text or "")
      lg.setColor(255,255,255, 255*math.abs(math.sin(GUI.sin_timer)))
      lg.line(wid+self.x+8,self.y+self.h-5, wid+self.x+20, self.y+self.h-5)
      lg.setColor(255,255,255)
   end
   return self
end

function TComboBox:onClick()
   if not self.readonly then
      GUI.focused_input = self
      GUI.text_buff = (GUI.focused_input.text or "")
      if love.system.getOS()=="Android" then love.keyboard.setTextInput(true) end
   end
end

------------------------------------------------------------------------------------

------------------------------------------------------------------------------------

local TDatePicker = {type = "TDatePicker", ancestor = "TLabel"}

TDatePicker.__index = TDatePicker

setmetatable(TDatePicker, {
  __index = TLabel, -- this is what makes the inheritance work
  __call = function (cls, ...)
    local self = setmetatable({}, cls)
    self:_init(...)
    return self
  end,
})


function TDatePicker:_init(x,y,w,h, init_data)
  TLabel._init(self, x,y,w,h, init_data) -- call the base class constructor
  self.type = "TDatePicker"
  self.ancestor = "TLabel"
  --self.bgcol = {125,025,055, 200}
  --self.h = {...}--x,y,w,h, init_data
  self.h = h
  self.fontColor = {255,255,255}
  -------------------------------------
  function self:onClick()
     if self.panl.visible then
        self.panl:hide()
     else
        self.panl:show()
     end
  end

 -- self.pane = GUI.newPanel(0,self.h, 7*34, 16+7*34):setParent(self):show():hide()
  local x = math.ceil((self.w - 7*34)/2)
  self.panl = GUI.newPanel(x,self.h, 7*34+2, 16+7*34+1):setParent(self)
  self.plbl = GUI.newLabel(0,0, 7*34, 33):setText(""):setParent(self.panl)
  self.plbl.bgcol = {200,200,200}
  self.panl.fill = true
  self.panl.topmost = true
  self.mmin = GUI.newButton(0,0,36,36):setParent(self.plbl):setText("<","center","center"):show()
  self.mmax = GUI.newButton(7*34-34,0,34,34):setParent(self.plbl):setText(">","center","center"):show()

  --self.pane.autosize=true
  self.panl.borders = true
  --self.pane.name = "test"
  self.date  = os.time()
  self.day   = tonumber(os.date("%d",self.date))
  self.month = tonumber(os.date("%m",self.date))
  self.year  = tonumber(os.date("%Y",self.date))
  self:setText(os.date("%d.%m.%Y",os.time({year=self.year, month=self.month, day = self.day, hour=0})),"center","center")


  self.butts = GUI.newGrid(1,34,7*34,7*34):setParent(self.panl):setSize(7,7):hide()
  self.butts.hightlightOnHover = true


  self.dows = {"Пн","Вт","Ср","Чт","Пт","Сб","Вс"}
  self.dowf = {"Понедельник", "Вторник", "Среда", "Четверг", "Пятница", "Суббота", "Воскресенье"}
  self.mnth = {"Январь", "Февраль", "Март", "Апрель", "Май", "Июнь", "Июль", "Август", "Сентябрь", "Октябрь", "Ноябрь", "Декабрь"}
  self.days_in_month = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }

  self.plbl:setText(self.mnth[self.month].. " " .. self.year,"center","top")


  --self.topmost = true

  self.butts.topmost = true
  self.mmin.topmost = true
  self.mmax.topmost = true

  function self.mmin.onClick()
     if self.year==self.syear and self.month<=self.smonth then return end
     self.month = self.month - 1

     if self.month < 1 then self.month = 12 self.year = self.year-1 end
     self.update_visuals()
     --print(self.month)
  end

  self.mmax.onClick = function()
     if self.year==self.eyear and self.month>=self.emonth then return end
     self.month = self.month + 1
     if self.month > 12 then self.month = 1 self.year = self.year+1 end
     self.update_visuals()
     --print(self.month)
  end


  function self.is_leap_year(year)
    return year % 4 == 0 and (year % 100 ~= 0 or year % 400 == 0)
  end

  function self.get_days_in_month(month, year)
    return month == 2 and self.is_leap_year(year) and 29
           or ("\31\28\31\30\31\30\31\31\30\31\30\31"):byte(month)
  end

  function self.butts.onClick(d,x,y)
     local _x, _y = d:getCellAt(x,y)
     --print(_x,_y)
     local day = tonumber(d.cells[_x][_y].text[1])

     if day then
        --print(day)
        self:setDate(os.time({year=self.year, month=self.month, day = day, hour=0}))
        --self.day = day
        --self:setText()
     end
     self.panl:hide()
  end

  function self.update_visuals()
      local td = tonumber(os.date("%w",os.time({year=self.year, month=self.month, day = 1, hour=0})))
      for i=1,7 do
          for j = 1,8 do
              if j==1 then
                 self.butts:setCellData(i,1, self.dows[i], {34,16},nil,nil,nil,{150,150,150})
              else
                 self.butts:setCellData(i, j, " ", {34,34},nil,nil,nil,{180,180,180})
                 self.butts.cells[i][j].bgcol = {180,180,180}
              end
          end
      end

      if td == 0 then td = 7 end
      local row = 2
      local col = td
      local day = 0
      local max_days = self.get_days_in_month(self.month, self.year)
      while day < max_days do
         day = day + 1
         local td = tonumber(os.date("%w",os.time({year=self.year, month=self.month, day = day, hour=0})))
         self.butts:setCellData(col,row, day, {34,34},nil,nil,nil,{220,220,220})
         self.butts.cells[col][row].bgcol = {220,220,220}
         if td == 0 then td = 7 end
         if td == 7 then row = row+1 end
         col = col+1
         if col>7 then col = 1 end
      end
      self.plbl:setText(self.mnth[self.month].. " " .. self.year,"center","top")
  end
  self.update_visuals()
  self.panl:hide()
  return self
end

function TDatePicker:getDate()
   return self.date
end

function TDatePicker:setDate(date)
   self.date  = date
   self.day   = tonumber(os.date("%d",self.date))
   self.month = tonumber(os.date("%m",self.date))
   self.year  = tonumber(os.date("%Y",self.date))
   self.update_visuals()
   self:setText(os.date("%d.%m.%Y",os.time({year=self.year, month=self.month, day = self.day, hour=0})),"center","center")
   return self
end

function TDatePicker:setDateBounds(sdate,edate)
   self.sdate = sdate
   self.sday   = tonumber(os.date("%d",self.sdate))
   self.smonth = tonumber(os.date("%m",self.sdate))
   self.syear  = tonumber(os.date("%Y",self.sdate))

   self.edate = edate
   self.eday   = tonumber(os.date("%d",self.edate))
   self.emonth = tonumber(os.date("%m",self.edate))
   self.eyear  = tonumber(os.date("%Y",self.edate))
   self.update_visuals()
   return self
end
--[[


function TDatePicker:onDraw()
   if not TControl.onDraw(self) then return end
   lg.setColor(self.bgColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("fill", self.x,self.y, self.w, self.h, self.curved_borders)
   lg.setColor(self.fontColor)
   if self.transparency then
      local col = {lg.getColor()}
      col[4] = self.transparency
      lg.setColor(col)
   end
   lg.rectangle("line", self.x,self.y, self.w, self.h, self.curved_borders)
   lg.print(self.text or "",self.x+5,self.y-5)
   if GUI.focused_input == self then
      local wid = lg.getFont():getWidth(self.text or "")
      lg.setColor(255,255,255, 255*math.abs(math.sin(GUI.sin_timer)))
      lg.line(wid+self.x+8,self.y+self.h-5, wid+self.x+20, self.y+self.h-5)
      lg.setColor(255,255,255)
   end
   return self
end ]]


------------------------------------------------------------------------------------

------------------------------------------------------------------------------------
GUI.available_controls = {}
GUI.available_controls["TControl"]      = TControl
GUI.available_controls["TImage"]        = TImage
GUI.available_controls["TPanel"]        = TPanel
GUI.available_controls["TLabel"]        = TLabel
GUI.available_controls["TButton"]       = TButton
GUI.available_controls["TStatsButton"]  = TStatsButton
GUI.available_controls["TScrollBox"]    = TScrollBox
GUI.available_controls["TScrollBar"]    = TScrollBar
GUI.available_controls["TEdit"]         = TEdit
GUI.available_controls["TProgressBar"]  = TProgressBar
GUI.available_controls["TCheckBox"]     = TCheckBox
GUI.available_controls["TPasswordEdit"] = TPasswordEdit
GUI.available_controls["TCombobox"]     = TComboBox
GUI.available_controls["TGrid"]         = TGrid
GUI.available_controls["TDatePicker"]   = TDatePicker
------------------------------------------------------------------------------------
GUI.newControl      = TControl
GUI.newImage        = TImage
GUI.newPanel        = TPanel
GUI.newLabel        = TLabel
GUI.newButton       = TButton
GUI.newStatsButton  = TStatsButton
GUI.newScrollBox    = TScrollBox
GUI.newScrollBar    = TScrollBar
GUI.newEdit         = TEdit
GUI.newProgressBar  = TProgressBar
GUI.newCheckBox     = TCheckBox
GUI.newPasswordEdit = TPasswordEdit
GUI.newComboBox     = TComboBox
GUI.newGrid         = TGrid
GUI.newDatePicker   = TDatePicker
------------------------------------------------------------------------------------


GUI.root = TControl(0, 0, GUI.width, GUI.height):show()
print(GUI.width, GUI.height)
GUI.font = lg.newFont(14)
lg.setFont(GUI.font)
GUI.debug_panel = GUI.newLabel(0,680,2000,40):setParent(GUI.root):setFontColor({255,255,255}):show()

return GUI

