if not app or not app.launched then
   SWAPPED = 'SWAPPED: false'
   app = {launched = false}
   app.classes = {}
   app.students = {}
   app.schedule = {}
   app.strings = {}
   app.buttons = {}
   
   love.filesystem.setIdentity("plus_plus",false)
   
   local filename_prefix = ""
   -- /data/user/0/org.love2d.android/files/save/archive
   -- /data/data/org.love2d.android/files/save/plus_plus
   if love.system.getOS()=='Android' then
      filename_prefix=love.filesystem.getSaveDirectory()..'/'
   end
   --love.window.showMessageBox( "debug", filename_prefix, "info", false)
   local version = {love.getVersion( )}
   --love.window.showMessageBox( "debug", table.concat(version,"."), "info", false)
   --love.window.showMessageBox( "debug", love.filesystem.getSaveDirectory( ), "info", false)
   
   local info = love.filesystem.getInfo("data.lua")
   if not info then
      love.filesystem.write("data.lua", "return {}" )
   end
   
   local _ = love.filesystem.load("data.lua")   
   local data = _() or {}   
   
   app.classes  = data.classes or app.classes
   app.students = data.students or app.students
   app.schedule = data.schedule or app.schedule
   app.data = {classes = app.classes, students = app.students, schedule = app.schedule}
end

function app:load_module(name)
   app.modules = app.modules or {}
   local _ = require(name)
   app.modules[name] = _
   for k,v in pairs(_) do
       app[k]=v
   end
end

function app:load_locale(locale)
   local f = love.filesystem.load('locales/'..locale..".lua")
   
   for k,v in pairs(f()) do
       app.strings[k] = v or k or '?'
   end
   local mt = {__index = function () return "undefined" end}
   setmetatable(app.strings, mt)
end


function app:rebuildGui(form)
   app.gui.controls = {}
   app.gui.root.children = {}   
   app.set_state(form)
end

function app:export_csv(filename, add_report)
end

function app:add_lesson(class, date)
end

function app:load_schedule(filename)
end

function app:add_scheduled_lesson(title, date, time)
end

function app:add_person(name, surname, class)
   local person = {name = name, surname = surname, marks={}, presense={}}
   self.classes[class][#app.classes[class]+1] = person
   self.people[#app.people+1] = person
end

function app:make_main_screen()
   local l = [[
   I need several screens/buttons:
     + main
       + classes (list all classes + number of people + if there's lesson today)
         + people (should be able to add those while on a class screen)
         + additional holidays per class (mb due to some event or planning)
       + schedule (set DAYs OF WEEK when there should be lessons, can only add anything if there's classes ready)
         + holidays
   ]]
   self.buttons.ClassesMain  = self.gui.newButton(math.ceil((self.gui.width-280)/2),80, 280,40, {root = true}):setText(self.strings['Classes'], 'center', 'center'):show()
   self.buttons.ScheduleMain = self.gui.newButton(math.ceil((self.gui.width-280)/2),160,280,40, {root = true}):setText(self.strings['Schedule'], 'center', 'center'):show()
   self.buttons.BackMain     = self.gui.newButton(math.ceil((self.gui.width-280)/2),app.gui.height-80,280,40,  {root = true}):setText(app.strings['Exit'], 'center', 'center'):show()
      
   function self.buttons.ClassesMain:onClick()
      app:rebuildGui("classes")
   end
   function self.buttons.ScheduleMain:onClick()
      app:rebuildGui("schedule")      
   end

   function self.buttons.BackMain:onClick()
      love.event.quit()
   end


end

l=[[
there are classes which contain people.
each person have marks and presence.
to build a full representation of a journal I need to traverse all persons in a class
and see if there's any data.
I'm to create a separate _tmp table with the DATE as a key, so I can sort that later (will need additional method for that)

Ideally, I'd like the app to automatically switch to a current day and let me choose the class I'm in today.
But I should be able to choose the date too.

Then again, I need to have some means to determine a schedule.
To add days etc.

Oh, and thin should work in portrait and scale up.
]]

local initialized = false
function init()
   app:make_main_screen()
end

function love.keyreleased(key)
   if key == "escape" then
      love.event.quit()
   end
end

function love.load()   
   app:load_locale('ru')
   app:load_module("helpers")
   app:load_module("states")
   app.gui = require("gui")
   app.gui.font = love.graphics.newFont('fonts/Roboto-Light.ttf',14)
   app.gui:set_rounded_corners(4)
   app.launched = true      
end

function love.update(dt)
   if not initialized then
      init()
      initialized = true
      return 
   end
   app.gui:update(dt)
end

function love.draw()
   app.gui:draw()
   love.graphics.setColor(255,255,255)
   local w,h,f = love.window.getMode( )
   love.graphics.print('Resolution: '..w..'x'..h, 10, 10)
   love.graphics.print('Flags: ',10, 30)
   local i = 0
   for k,v in pairs(f) do       
       love.graphics.print(k..' = '.. tostring(v), 10+love.graphics.getFont():getWidth('Flags: '), 30+i*20)
       i = i + 1
   end
   love.graphics.print(app.gui.width, 10+love.graphics.getFont():getWidth('Flags: '), 30+(i+2)*20)
   love.graphics.print(app.gui.height, 10+love.graphics.getFont():getWidth('Flags: '), 30+(i+3)*20)
   love.graphics.print(SWAPPED, 10+love.graphics.getFont():getWidth('Flags: '), 30+(i+4)*20)
end